from cayenne.simulation import Simulation
# import matplotlib.pyplot as plt

model_str = """
const compartment comp1;
comp1 = 1.0; # volume of compartment

activation_a        : Da + A    => Dpa      ; gammaA;
activation_r        : Dr + A    => Dpr      ; gammaR;
deactivation_a      : Dpa       => Da + A   ; thetaA;
deactivation_r      : Dpr       => Dr + A   ; thetaR;
transcription_a     : Da        => Da + Ma  ; alphaA;
transcription_r     : Dr        => Dr + Mr  ; alphaR;
transcription_p_a   : Dpa       => Dpa + Ma ; alphapA;
transcription_p_r   : Dpr       => Dpr + Mr ; alphapR;
translation_a       : Ma        => Ma + A   ; betaA;
translation_r       : Mr        => Mr + R   ; betaR;
complexation        : A + R     => C        ; gammaC;
decomplexation      : C         => R        ; gammaA;
decay_mRNA_a        : Ma        =>          ; deltaMA;
decay_mRNA_r        : Mr        =>          ; deltaMR;
decay_prot_a        : A         =>          ; deltaA;
decay_prot_r        : R         =>          ; deltaR;

alphaA = 50;
alphapA = 500;
alphaR = 0.01;
alphapR = 50;
betaA = 50;
betaR = 5;
deltaMA = 10;
deltaMR = 0.5;
deltaA = 1;
deltaR = 0.2;
gammaA = 1;
gammaR = 2;
gammaC = 2;
thetaA = 50;
thetaR = 100;
chem_flag = false;

Da = 1;
Dr = 1;
Dpa = 0;
Dpr = 0;
Ma = 0;
Mr = 0;
A = 0;
R = 0;
C = 0;
"""
sim = Simulation.load_model(model_str, "ModelString")
# Run the simulation
sim.simulate(max_t=200, max_iter=700000, algorithm='direct')
print(sim.results)
# sim.plot()
# plt.show()



import numpy as np
import numpy.testing as npt
import pytest
import rebop
import xarray as xr


def sir_model(transmission: float = 1e-4, recovery: float = 0.01) -> rebop.Gillespie:
    sir = rebop.Gillespie()
    sir.add_reaction(transmission, ["S", "I"], ["I", "I"])
    sir.add_reaction(recovery, ["I"], ["R"])
    return sir


@pytest.mark.parametrize("seed", [None, *range(10)])
def test_sir(seed: int | None) -> None:
    sir = sir_model()
    ds = sir.run({"S": 999, "I": 1}, tmax=250, nb_steps=250, rng=seed)
    assert isinstance(ds, xr.Dataset)
    npt.assert_array_equal(ds.time, np.arange(251))
    assert all(ds.S >= 0)
    assert all(ds.I >= 0)
    assert all(ds.R >= 0)
    assert all(ds.S <= 999)
    assert all(ds.I <= 1000)
    assert all(ds.R <= 1000)
    npt.assert_array_equal(ds.S + ds.I + ds.R, [1000] * 251)


def test_fixed_seed() -> None:
    sir = sir_model()
    ds = sir.run({"S": 999, "I": 1}, tmax=250, nb_steps=250, rng=42)

    assert ds.S[-1] == 0
    assert ds.I[-1] == 227
    assert ds.R[-1] == 773


@pytest.mark.parametrize("seed", range(10))
def test_all_reactions(seed: int) -> None:
    tmax = 250
    sir = sir_model()
    ds = sir.run({"S": 999, "I": 1}, tmax=tmax, nb_steps=0, rng=seed)
    assert ds.time[0] == 0
    assert ds.time[-1] > tmax
    assert all(ds.time.diff(dim="time") > 0)
    if np.isinf(ds.time[-1].to_numpy()):
        ds = ds.isel(time=range(ds.time.size - 1))
    dds = ds.diff(dim="time")
    assert set(dds.S.to_numpy()) <= {-1, 0}
    assert set(dds.I.to_numpy()) <= {-1, 1}
    assert set(dds.R.to_numpy()) <= {0, 1}


def test_dense_vs_sparse() -> None:
    sir = sir_model()
    init = {"S": 999, "I": 1}
    tmax = 250
    nb_steps = 250
    seed = 42
    ds_auto = sir.run(init, tmax=tmax, nb_steps=nb_steps, rng=seed, sparse=None)
    ds_dense = sir.run(init, tmax=tmax, nb_steps=nb_steps, rng=seed, sparse=False)
    ds_sparse = sir.run(init, tmax=tmax, nb_steps=nb_steps, rng=seed, sparse=True)
    assert (ds_dense == ds_auto).all()
    assert (ds_dense == ds_sparse).all()


@pytest.mark.parametrize("nb_steps", [0, 250])
def test_var_names(nb_steps: int) -> None:
    all_variables = {"S", "I", "R"}
    subset_to_save = ["S", "I"]
    remaining = all_variables.difference(subset_to_save)

    sir = sir_model()
    init = {"S": 999, "I": 1}
    tmax = 250
    seed = 0

    ds_all = sir.run(init, tmax=tmax, nb_steps=nb_steps, rng=seed, var_names=None)
    ds_subset = sir.run(
        init, tmax=tmax, nb_steps=nb_steps, rng=seed, var_names=subset_to_save
    )

    for s in subset_to_save:
        assert s in ds_subset
    for s in remaining:
        assert s not in ds_subset

    assert ds_all[subset_to_save] == ds_subset


def test_arbitrary_rates() -> None:
    s = rebop.Gillespie()
    with pytest.raises(ValueError, match="Rate expression not understood"):
        s.add_reaction("+", [], ["A"])
    with pytest.raises(ValueError, match="Rate expression not understood"):
        s.add_reaction("1+", [], ["A"])

    s = rebop.Gillespie()
    s.add_reaction("B", [], ["A"])
    with pytest.warns(
        UserWarning,
        match="species are not involved in any reaction",
    ):
        ds = s.run({"B": 1}, tmax=10, nb_steps=100)
    assert ds.A[-1] >= 1
    np.testing.assert_equal(ds.B.values, [1] * 101)

    s = rebop.Gillespie()
    s.add_reaction("B", [], ["A"])
    s.add_reaction(1, [], ["B"])
    ds_with_b = s.run({}, tmax=10, nb_steps=100)

    assert ds_with_b.A[-1] >= 1


def test_arbitrary_rates_crossed() -> None:
    s = rebop.Gillespie()
    s.add_reaction("B", [], ["A"])
    s.add_reaction("A", [], ["B"])
    ds = s.run({}, tmax=10, nb_steps=10)
    expected = xr.Dataset(
        {"A": ("time", [0] * 11), "B": ("time", [0] * 11)},
        coords={"time": np.linspace(0, 10, 11)},
    )
    xr.testing.assert_identical(ds, expected)

    ds = s.run({"A": 1}, tmax=10, nb_steps=10)
    assert ds.A[-1] > 1
    assert ds.B[-1] > 0

    ds = s.run({"B": 1}, tmax=10, nb_steps=10)
    assert ds.A[-1] > 0
    assert ds.B[-1] > 1


def test_arbitrary_rates_2() -> None:
    s = rebop.Gillespie()
    s.add_reaction(14, [], ["A"])
    s.add_reaction(0.1, ["A", "B"], ["C"], 0.01)
    s.add_reaction("0.2 * B * C / (5 + C)", ["B"], ["D"])
    ds = s.run({"B": 1000}, tmax=100, nb_steps=100)

    assert (ds.B + ds.C + ds.D == 1000).all()
    assert ds.D[-1] >= 1


def test_run_empty() -> None:
    s = rebop.Gillespie()
    ds = s.run({}, tmax=10, nb_steps=10)
    expected = xr.Dataset(coords={"time": np.linspace(0, 10, 11)})
    xr.testing.assert_identical(ds, expected)


def test_parameters() -> None:
    s = rebop.Gillespie()
    s.add_reaction(4, ["A"], ["B"])
    with pytest.raises(ValueError, match="Species B cannot also be a parameter"):
        s.run({}, 10, 10, params={"B": 4.2})

    s = rebop.Gillespie()
    s.add_reaction("k", [], ["A"])
    with pytest.raises(ValueError, match="Parameter k should have a value"):
        s.run({}, 10, 10)
    ds = s.run({}, 10, 10, params={"k": 0.4})
    assert ds.A[-1] > 0



"""Demonstration of a non LMA rate and of rate parameters.

The reaction here is A → B with the enzymatic rate V * A / (Km + A).

V and Km are provided as parameters in the `run` method.
"""

import rebop

mm = rebop.Gillespie()
mm.add_reaction("V * A / (Km + A)", ["A"], ["P"])

print(mm)

params = {"V": 1, "Km": 20}

ds = mm.run({"A": 100}, tmax=250, nb_steps=100, params=params, rng=0)

print(ds)



"""The classical SIR model.

There are three categories of people:
* S: Susceptible
* I: Infected
* R: Recovered

    S + I -> 2 I
    I     ->   R
"""

import rebop

sir = rebop.Gillespie()
sir.add_reaction(1e-4, ["S", "I"], ["I", "I"])
sir.add_reaction(0.01, ["I"], ["R"])

print(sir)

ds = sir.run({"S": 999, "I": 1}, tmax=250, nb_steps=250, rng=0)

print(ds)



import rebop

alphaA = 50
alphapA = 500
alphaR = 0.01
alphapR = 50
betaA = 50
betaR = 5
deltaMA = 10
deltaMR = 0.5
deltaA = 1
deltaR = 0.2
gammaA = 1
gammaR = 1
gammaC = 2
thetaA = 50
thetaR = 100

vilar = rebop.Gillespie()
vilar.add_reaction(gammaA, ['Da', 'A'], ['Dpa'])
vilar.add_reaction(gammaR, ['Dr', 'A'], ['Dpr'])
vilar.add_reaction(thetaA, ['Dpa'], ['Da', 'A'])
vilar.add_reaction(thetaR, ['Dpr'], ['Dr', 'A'])
vilar.add_reaction(alphaA, ['Da'], ['Da', 'Ma'])
vilar.add_reaction(alphaR, ['Dr'], ['Dr', 'Mr'])
vilar.add_reaction(alphapA, ['Dpa'], ['Dpa', 'Ma'])
vilar.add_reaction(alphapR, ['Dpr'], ['Dpr', 'Mr'])
vilar.add_reaction(betaA, ['Ma'], ['Ma', 'A'])
vilar.add_reaction(betaR, ['Mr'], ['Mr', 'R'])
vilar.add_reaction(gammaC, ['A', 'R'], ['C'])
vilar.add_reaction(deltaA, ['C'], ['R'])
vilar.add_reaction(deltaMA, ['Ma'], [])
vilar.add_reaction(deltaMR, ['Mr'], [])
vilar.add_reaction(deltaA, ['A'], [])
vilar.add_reaction(deltaR, ['R'], [])

print(vilar)

for _ in range(100):
    ds = vilar.run({'Da': 1, 'Dr': 1}, tmax=200, nb_steps=200)



"""rebop is a fast stochastic simulator for well-mixed chemical reaction networks."""

from __future__ import annotations

import warnings
from collections.abc import Mapping, Sequence

import numpy as np
import xarray as xr

from rebop import _lib

__all__ = ("Gillespie", "__version__")


__version__: str = _lib.__version__

type SeedLike = int | np.integer | Sequence[int] | np.random.SeedSequence
type RNGLike = np.random.Generator | np.random.BitGenerator


class Gillespie:
    """Reaction system composed of species and reactions."""

    def __init__(self) -> None:
        """Initialize a solver."""
        self.gillespie = _lib.Gillespie()

    def add_reaction(
        self,
        rate: float | str,
        reactants: Sequence[str],
        products: Sequence[str],
        reverse_rate: float | str | None = None,
    ) -> None:
        """Add a reaction to the system.

        Reaction rates can be specified with a number (for a reaction obeying
        the Law of Mass Action) or a string (for an arbitrary reaction rate).
        The string can involve any species involved in reactions, and also
        parameters defined with the `params` argument of the `run` method.

        If you can, use the law of mass action, which is probably going to be
        more efficient and can be more correct. For example, for a dimerisation
        equation:

        ```python
        s = Gillespie()
        # Correct, and recommended
        s.add_reaction(4.2, ["A", "A"], ["AA"])
        # Correct, but not recommended
        s.add_reaction("4.2 * A * (A-1)", ["A", "A"], ["AA"])
        # Incorrect (this would be a constant propensity)
        s.add_reaction("4.2", ["A", "A"], ["AA"])
        # Incorrect (incorrect expression)
        s.add_reaction("4.2 * A^2", ["A", "A"], ["AA"])
        ```

        Example
        -------
        ```python
        s = Gillespie()
        # Add the birth equation ø -> A at rate 14
        s.add_reaction(14, [], ["A"])
        # Add the reversible reaction A + B -> C
        # with forward rate 0.1 and reverse rate 0.01
        s.add_reaction(0.1, ["A", "B"], ["C"], 0.01)
        # Add the transformation B -> C with a non-LMA rate
        s.add_reaction("2.1 * B * A / (5 + A)", ["B"], ["C"])
        # Add the transformation B -> C with a non-LMA rate and parameters
        # that need to be defined by passing `params={"k": 2.1, "K_A": 5}` to
        # the method `run`
        s.add_reaction("k * B * A / (K_A + A)", ["B"], ["C"])
        ```

        Parameters
        ----------
        rate : float | str
            If numeric, this is the rate constant of a Law of Mass Action.
            If string, this is the mathematical expression of the reaction rate.
        reactants : Sequence[str]
            List of the species that are on the left-hand-side of the reaction.
        products : Sequence[str]
            List of the species that are on the right-hand-side of the reaction.
        reverse_rate : float | str | None
            Rate of the reverse reaction, if this reaction is reversible.
            This is just a convenience to avoid using `add_reaction` a second
            time for the reverse reaction.
        """
        self.gillespie.add_reaction(rate, reactants, products, reverse_rate)

    def __str__(self) -> str:
        """Return a textual representation of the reaction system."""
        return str(self.gillespie)

    def run(  # noqa: PLR0913 too many parameters in function definition
        self,
        init: Mapping[str, int],
        tmax: float,
        nb_steps: int,
        *,
        params: Mapping[str, float] | None = None,
        rng: RNGLike | SeedLike | None = None,
        sparse: bool | None = None,
        var_names: Sequence[str] | None = None,
    ) -> xr.Dataset:
        """Run the system until `tmax` with `nb_steps` steps.

        Parameters
        ----------
        init : Mapping[str, int]
            Dictionary that indicates the initial condition. It can omit
            species, their initial number will be 0.
        tmax : float
            Simulation end time.
        nb_steps : int
            Number of steps to return, equally distributed between 0 and `tmax`.
            If 0, then all reactions are returned.
        params : Mapping[str, float] | None
            Dictionary of values for the parameters that appear in the rates.
        rng : RNGLike | SeedLike | None
            Numpy `Generator`, `BitGenerator` or seed.
        sparse : bool | None
            Whether to internally represent reactions as dense or sparse.
            Sparse reactions will tend to be faster for systems with many
            species. If `None`, rebop will choose depending on heuristics.
        var_names : Sequence[str] | None
            List of variable names to return, if one does not want to save
            all variables. If `None`, all variables will be saved.

        Returns
        -------
        xr.Dataset
            `xarray.Dataset` object with a `time` dimension and variables
            being chemical species (all of them, or just the subset defined
            by `var_names`).
        """
        params = {} if params is None else params
        rng_ = np.random.default_rng(rng)
        seed = rng_.integers(np.iinfo(np.uint64).max, dtype=np.uint64)
        try:
            self.gillespie.set_init(init)
        except UserWarning as e:
            warnings.warn(e, stacklevel=2)
        times, result = self.gillespie.run(
            tmax,
            nb_steps,
            params=params,
            seed=seed,
            sparse=sparse,
            var_names=var_names,
        )
        ds = xr.Dataset(
            data_vars={
                name: xr.DataArray(values, dims="time", coords={"time": times})
                for name, values in result.items()
            },
            coords={"time": times},
        )
        return ds



import gillespy2
import numpy as np
import matplotlib.pyplot as plt

class VilarOscillator(gillespy2.Model):  # type: ignore[misc]
    def __init__(self, parameter_values: None = None) -> None:
        gillespy2.Model.__init__(self, name="VilarOscillator")
        self.volume = 1

        # Parameters
        alphaA = gillespy2.Parameter(name="alphaA", expression=50)
        alphaA_prime = gillespy2.Parameter(name="alphaA_prime", expression=500)
        alphaR = gillespy2.Parameter(name="alphaR", expression=0.01)
        alphaR_prime = gillespy2.Parameter(name="alphaR_prime", expression=50)
        betaA = gillespy2.Parameter(name="betaA", expression=50)
        betaR = gillespy2.Parameter(name="betaR", expression=5)
        deltaMA = gillespy2.Parameter(name="deltaMA", expression=10)
        deltaMR = gillespy2.Parameter(name="deltaMR", expression=0.5)
        deltaA = gillespy2.Parameter(name="deltaA", expression=1)
        deltaR = gillespy2.Parameter(name="deltaR", expression=0.2)
        gammaA = gillespy2.Parameter(name="gammaA", expression=1)
        gammaR = gillespy2.Parameter(name="gammaR", expression=1)
        gammaC = gillespy2.Parameter(name="gammaC", expression=2)
        thetaA = gillespy2.Parameter(name="thetaA", expression=50)
        thetaR = gillespy2.Parameter(name="thetaR", expression=100)

        self.add_parameter([alphaA, alphaA_prime, alphaR, alphaR_prime, betaA, betaR,
                            deltaMA, deltaMR, deltaA, deltaR, gammaA, gammaR, gammaC,
                            thetaA, thetaR])

        # Species
        Da = gillespy2.Species(name="Da", initial_value=1, mode="discrete")
        Da_prime = gillespy2.Species(name="Da_prime", initial_value=0, mode="discrete")
        Ma = gillespy2.Species(name="Ma", initial_value=0, mode="discrete")
        Dr = gillespy2.Species(name="Dr", initial_value=1, mode="discrete")
        Dr_prime = gillespy2.Species(name="Dr_prime", initial_value=0, mode="discrete")
        Mr = gillespy2.Species(name="Mr", initial_value=0, mode="discrete")
        C = gillespy2.Species(name="C", initial_value=0, mode="discrete")
        A = gillespy2.Species(name="A", initial_value=0, mode="discrete")
        R = gillespy2.Species(name="R", initial_value=0, mode="discrete")

        self.add_species([Da, Da_prime, Ma, Dr, Dr_prime, Mr, C, A, R])

        # Reactions
        r1 = gillespy2.Reaction(name="r1", reactants={'A': 1, 'R': 1}, products={'C': 1}, rate="gammaC")
        r2 = gillespy2.Reaction(name="r2", reactants={'A': 1}, products={}, rate="deltaA")
        r3 = gillespy2.Reaction(name="r3", reactants={'C': 1}, products={'R': 1}, rate="deltaA")
        r4 = gillespy2.Reaction(name="r4", reactants={'R': 1}, products={}, rate="deltaR")
        r5 = gillespy2.Reaction(name="r5", reactants={'A': 1, 'Da': 1}, products={'Da_prime': 1}, rate="gammaA")
        r6 = gillespy2.Reaction(name="r6", reactants={'Da_prime': 1}, products={'A': 1, 'Da': 1}, rate="thetaA")
        r7 = gillespy2.Reaction(name="r7", reactants={'Da': 1}, products={'Da': 1, 'Ma': 1}, rate="alphaA")
        r8 = gillespy2.Reaction(name="r8", reactants={'Da_prime': 1}, products={'Da_prime': 1, 'Ma': 1}, rate="alphaA_prime")
        r9 = gillespy2.Reaction(name="r9", reactants={'Ma': 1}, products={}, rate="deltaMA")
        r10 = gillespy2.Reaction(name="r10", reactants={'Ma': 1}, products={'A': 1, 'Ma': 1}, rate="betaA")
        r11 = gillespy2.Reaction(name="r11", reactants={'A': 1, 'Dr': 1}, products={'Dr_prime': 1}, rate="gammaR")
        r12 = gillespy2.Reaction(name="r12", reactants={'Dr_prime': 1}, products={'A': 1, 'Dr': 1}, rate="thetaR")
        r13 = gillespy2.Reaction(name="r13", reactants={'Dr': 1}, products={'Dr': 1, 'Mr': 1}, rate="alphaR")
        r14 = gillespy2.Reaction(name="r14", reactants={'Dr_prime': 1}, products={'Dr_prime': 1, 'Mr': 1}, rate="alphaR_prime")
        r15 = gillespy2.Reaction(name="r15", reactants={'Mr': 1}, products={}, rate="deltaMR")
        r16 = gillespy2.Reaction(name="r16", reactants={'Mr': 1}, products={'Mr': 1, 'R': 1}, rate="betaR")

        self.add_reaction([r1, r2, r3, r4, r5, r6, r7, r8, r9,
                           r10, r11, r12, r13, r14, r15, r16])

        # Timespan
        self.timespan(np.linspace(0, 200, 201))

model = VilarOscillator()
results = model.run(gillespy2.SSACSolver, number_of_trajectories=90)
print(results)
# results.plot()
# plt.show()



import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as stats

with open('vilar.csv', 'r') as file:
    header = file.readline().rstrip()[1:].split(',')[1:]

data = np.loadtxt('vilar.csv', delimiter=',', skiprows=1)
ns = data[:, 0]

for i in [3, 1, 2, 4, 0]:
    name = header[i]
    times = data[:, i+1]
    res = stats.linregress(ns, times)
    plt.plot(ns, times, 'o', color=f'C{i}', label=name)
    plt.plot(ns, res.intercept + res.slope * ns,
             label=f'{res.intercept:.3f} + {res.slope:.3f} * N', color=f'C{i}')

plt.xlabel('Number of trajectories')
plt.ylabel('Time [s]')
plt.title('Vilar oscillator')
plt.grid(True)
plt.legend()
plt.savefig('vilar.png', bbox_inches='tight')
# plt.show()
